; complex.lsp

(defun make-complex (re im)
    (cond
	((eql im 0) re)
	(t (make-array 2 :initial-contents (list re im))) ) )

(defun c+ (&rest b)
    (let ((resum 0) (imsum 0) (re) (im))
	(loop
	    (if (atom b) (return))
	    (cond
		((arrayp (car b))
		 (setq re (aref (car b) 0))
		 (setq im (aref (car b) 1))
		 (setq resum (+ resum re))
		 (setq imsum (+ imsum im)) )
		(t (setq resum (+ resum (car b)))) )
	    (setq b (cdr b)) )
	(make-complex resum imsum) ) )

(defun c- (&rest b)
    (cond
	((not (consp b)) (print '(bad usage)) (return nil))
	((atom (cdr b)) (c* -1 (car b)))
	(t (let ((re) (im))
	    (cond
		((not (arrayp (car b))) (setq re (car b)) (setq im 0))
		(t
		 (setq re (aref (car b) 0))
		 (setq im (aref (car b) 1)) ) )
	    (setq b (cdr b))
	    (loop
		(if (atom b) (return (make-complex re im)))
		(cond
		    ((arrayp (car b))
		     (setq re (- re (aref (car b) 0)))
		     (setq im (- im (aref (car b) 1))) )
		    (t
		     (setq re (- re (car b))) ) )
		(setq b (cdr b)) ) ) ) ) )

(defun c* (&rest b)
    (let ((re 1) (im 0) (x) (y) (newre) (newim))
	(loop
	    (if (atom b) (return))
	    (cond
		((not (arrayp (car b)))
		 (setq re (* re (car b)))
		 (setq im (* im (car b))) )
		(t
		    (setq x (aref (car b) 0))
		    (setq y (aref (car b) 1))
		    (setq newre (- (* re x) (* im y)))
		    (setq newim (+ (* re y) (* im x)))
		    (setq re newre)
		    (setq im newim) ) )
	    (setq b (cdr b)) )
	(make-complex re im) ) )

(defun c/ (&rest b)
    (cond
	((not (consp b)) (print '(bad usage)) nil)
	((atom (cdr b)) (c/ 1 (car b)))
	(t
	 (let ((re) (im) (x) (y) (bo) (newre) (newim))
	    (cond
		((not (arrayp (car b)))
		 (setq re (car b))
		 (setq im 0) )
		(t
		 (setq re (aref (car b) 0))
		 (setq im (aref (car b) 1)) ) )
	    (setq b (cdr b))
	    (loop
		(if (atom b) (return (make-complex re im)))
		(cond
		    ((not (arrayp (car b)))
		     (setq x (car b))
		     (setq y 0) )
		    (t
		     (setq x (aref (car b) 0))
		     (setq y (aref (car b) 1)) ) )
		(setq bo (+ (* x x) (* y y)))
		(setq newre (+ (* re x) (* im y)))
		(setq newim (- (* im x) (* re y)))
		(setq re (/ newre bo))
		(setq im (/ newim bo))
		(setq b (cdr b)) ) ) ) ) )

(defun csqrt (z)
    (cond
	((and (not (arrayp z)) (minusp z))
	    (let ((re z) (im 0) (arg) (r))
		(setq arg (/ (complex-arg re im) 2))
		(setq r (expt (+ (* re re) (* im im)) 1/4))
		(make-complex (* r (cos arg)) (* r (sin arg))) ) )
	((not (arrayp z)) (sqrt z))
	(t
	    (let ((re (aref z 0)) (im (aref z 1)) (arg) (r))
		(setq arg (/ (complex-arg re im) 2))
		(setq r (expt (+ (* re re) (* im im)) 1/4))
		(make-complex (* r (cos arg)) (* r (sin arg))) ) ) ) )

(defun ccos (z)
    (cond
	((not (arrayp z)) (cos z))
	(t
	    (let ((x (aref z 0)) (y (aref z 1)))
		(make-complex
		    (/ (* (+ (exp y) (exp (* -1 y))) (cos x)) 2)
		    (/ (* (- (exp y) (exp (* -1 y))) (sin x)) -2) ) ) ) ) )

(defun csin (z)
    (let ((x) (y))
	(cond
	    ((not (arrayp z)) (sin z))
	    (t
		(setq x (aref z 0))
		(setq y (aref z 1))
		(make-complex (* (+ (exp y) (exp (* -1 y))) (sin x) 1/2)
			      (* (- (exp y) (exp (* -1 y))) (cos x) 1/2) ) ) ) ) )

(defun ctan (z)
    (c/ (csin z) (ccos z)) )

(defun cexp (z)
    (cond
	((not (arrayp z)) (exp z))
	(t
	    (let ((x (aref z 0)) (y (aref z 1)) (a))
		(setq a (exp x))
		(make-complex (* a (cos y)) (* a (sin y))) ) ) ) )

(defun cexpt (z1 z2)
    (let ((r1) (a1) (x1) (y1) (x2) (y2) (r3) (a3))
	(cond
	    ((eql z1 0)
	     (cond
		((eql z2 0) 1)
		((eql z2 0.0) (print '(divided by zero)) nil)
		(t 0) ) )
	    ((eql z1 0.0)
	     (cond
		((eql z2 0) 1)
		((eql z2 0.0) (print '(divided by zero)) nil)
		(t 0.0) ) )
	    (t
	     (cond
		((and (not (arrayp z1)) (minusp z1)) (setq r1 (* -1 z1)) (setq a1 pi))
		((not (arrayp z1)) (setq r1 z1) (setq a1 0))
		(t
		    (setq x1 (aref z1 0))
		    (setq y1 (aref z1 1))
		    (setq r1 (sqrt (+ (* x1 x1) (* y1 y1))))
		    (setq a1 (complex-arg x1 y1)) ) )
	     (cond
		((not (arrayp z2)) (setq x2 z2) (setq y2 0))
		(t
		    (setq x2 (aref z2 0))
		    (setq y2 (aref z2 1)) ) )
	     (setq r3 (* (expt r1 x2) (exp (* -1 a1 y2))))
	     (setq a3 (+ (* x2 a1) (* y2 (log r1))))
	     (make-complex (* r3 (cos a3)) (* r3 (sin a3))) ) ) ) )

(defun clog (z)
    (let ((x) (y) (r) (a))
	(cond
	    ((and (not (arrayp z)) (minusp z)) (setq r (* -1 z)) (setq a pi) (make-complex (log r) a))
	    ((not (arrayp z)) (log z))
	    (t
		(setq x (aref z 0))
		(setq y (aref z 1))
		(setq r (sqrt (+ (* x x) (* y y))))
		(setq a (complex-arg x y))
		(make-complex (log r) a) ) ) ) )

(defun catan (z)
    (cond
	((not (arrayp z)) (atan z))
	(t
	 (c* (make-complex 0 1/2)
	    (clog (c/ (c* -1 (cexpt (c+ z (make-complex 0 1)) 2)) (c+ (c* z z) 1))) ) ) ) )

(defun cacos (z)
    (cond
	((not (arrayp z)) (acos z))
	(t
	 (c* (make-complex 0 -1)
	     (clog (c+ z (csqrt (c- (c* z z) 1)))) ) ) ) )

(defun casin (z)
    (cond
	((not (arrayp z)) (asin z))
	(t
	 (c* (make-complex 0 -1)
	     (clog (c+ (c* (make-complex 0 1) z) (csqrt (c- 1 (c* z z))))) ) ) ) )

(defun realpart (z)
    (aref z 0) )

(defun imagpart (z)
    (aref z 1) )

;-----------------------------------------------------------------------------------------------
(defun complex-arg (x y)
    (let ((arg))
	(cond
	    ((and (= x 0) (< 0 y)) (/ pi 2))
	    ((and (= x 0) (< y 0)) (/ pi -2))
	    ((and (= x 0) (= y 0)) (print '(i do not know the arg)) nil)
	    (t
		(setq arg (atan (/ y x)))
		(cond
		    ((and (< 0 x) (< 0 y)) arg)
		    ((and (< x 0) (< 0 y)) (+ arg pi))
		    ((and (< 0 x) (< y 0)) arg)
		    ((and (< x 0) (< y 0)) (- arg pi))
		    ((and (< 0 x) (eql y 0)) 0)
		    ((and (< 0 x) (eql y 0.0)) 0.0)
		    ((and (< x 0) (= y 0)) pi) ) ) ) ) )
